//
//   "F_Sexyloop.java"
//
//   F-Sexyloop simulator based on Sayama's Evoloop simulator (Copyright (c) 2004 by Hiroki Sayama)
//	 with event-driven birth/death detection mechanisms
//   Java application version 0.2
//
//   Final version created in 08/2008 by Oros Nicolas
//
//  Remark: some of the code is not used anymore...and some bits can be improved...I'll have to clean it a bit...


import java.awt.*;
import java.awt.event.*;
import java.awt.Graphics;
import java.io.*;
import java.util.*;
import java.lang.*;

// Canvas for graphical display of CA configuration

class CellularAutomataCanvas extends Canvas
{
    public CellularAutomata caller;
    public Image buffer = null;
    public Graphics buffer_g = null;
    public int canvasWidth, canvasHeight;

    CellularAutomataCanvas(CellularAutomata obj, int cx, int cy)
    {
		caller = obj;
		canvasWidth = cx;
		canvasHeight = cy;
		setSize(canvasWidth, canvasHeight);
		setBackground(Color.white);
    }

    public void createBuffer()
    {
		buffer = createImage(caller.width, caller.height);
		buffer_g = buffer.getGraphics();
		buffer_g.setColor(caller.cellColor[caller.QUIESCENT]);
		buffer_g.fillRect(0, 0, caller.width, caller.height);
    }

    public void paint(Graphics g)
    {
		if (buffer == null) createBuffer();
		g.drawImage(buffer, 0, 0, canvasWidth, canvasHeight, this);
    }

    public void update(Graphics g)
    {
		paint(g);
    }
}


// Main Frame, including CA updating scheme and event detection mechanisms

class CellularAutomata extends Frame implements ActionListener
{
    public int width = 200;
    public int height = 200;
    public int[][] cell;
    public int[][] scan;
    public StateChangeStack stateChanges;
    public CellUpdateStack cellUpdates;

    public SpeciesDatabase population;
    public BirthEventStack birthEvents;
    public DeathEventStack deathEvents;

    public static final int QUIESCENT = 0;
    public static final int NOT_OCCUPIED = -1;
    public static Color[] cellColor = {Color.black, Color.blue, Color.red,
				       Color.magenta, Color.green, Color.cyan,
				       Color.yellow, Color.white, Color.LIGHT_GRAY, Color.orange, Color.pink, Color.darkGray};

    public static Color[] scanColor =
    {
		new Color(255, 255, 128),
		new Color(255, 128, 255),
		new Color(255, 128, 128),
		new Color(128, 255, 255),
		new Color(128, 255, 128),
		new Color(128, 128, 255),
		new Color(255, 255, 192),
		new Color(255, 192, 255),
		new Color(255, 192, 192),
		new Color(255, 192, 128),
		new Color(255, 128, 192),
		new Color(192, 255, 255),
		new Color(192, 255, 192),
		new Color(192, 255, 128),
		new Color(192, 192, 255),
		new Color(192, 192, 192),
		new Color(192, 192, 128),
		new Color(192, 128, 255),
		new Color(192, 128, 192),
		new Color(192, 128, 128),
		new Color(128, 255, 192),
		new Color(128, 192, 255),
		new Color(128, 192, 192),
		new Color(128, 192, 128),
		new Color(128, 128, 192)
	};

    public Panel simulator;
    public CellularAutomataCanvas cnv;
    public Panel console;
    public Panel variousSettings;
    public Label timeDisplay;
    public TextField intervalField;
    public TextField spaceSizeField;
    public TextField ancestorField;
    public Panel buttons;
    public Button startStop, step, reset;
    public Panel analysisResultPanel;
    public TextArea analysisResult;

    public BufferedWriter out;
    public int nbReproSexArm, nbReproSexLoop, nbReproAsexualLoop;

    public boolean pause, oneStep, resetRequest;
    public long time;
    public int interval;

    CellularAutomata(int cx, int cy)
    {
		super("F-Sexyloop");

		addWindowListener(new WindowAdapter()
		{
			public void windowClosing(WindowEvent e)
			{
				System.exit(0);
			}
		});

		setLayout(new BorderLayout(10, 10));

		simulator = new Panel();
		simulator.setLayout(new BorderLayout(10, 10));
		simulator.add(cnv = new CellularAutomataCanvas(this, cx, cy), BorderLayout.CENTER);

		console = new Panel();
		console.setLayout(new GridLayout(2, 1));

		buttons = new Panel();
		buttons.setLayout(new GridLayout(1, 5));
		buttons.add(new Label("Time: ", Label.RIGHT));
		buttons.add(timeDisplay = new Label("0", Label.LEFT));
		buttons.add(startStop = new Button("Start"));
		startStop.addActionListener(this);
		buttons.add(step = new Button("Step"));
		step.addActionListener(this);
		buttons.add(reset = new Button("Reset"));
		reset.addActionListener(this);
		console.add(buttons);

		variousSettings = new Panel();
		variousSettings.setLayout(new GridLayout(1, 6));
		variousSettings.add(new Label("Interval: ", Label.RIGHT));
		variousSettings.add(intervalField = new TextField("100"));
		intervalField.addActionListener(this);
		variousSettings.add(new Label("Space Size: ", Label.RIGHT));
		variousSettings.add(spaceSizeField = new TextField("1000"));
		spaceSizeField.addActionListener(this);
		variousSettings.add(new Label("Ancestor: ", Label.RIGHT));
		variousSettings.add(ancestorField = new TextField("13-2-SEX.img"));
		ancestorField.addActionListener(this);
		console.add(variousSettings);

		simulator.add(console, BorderLayout.SOUTH);
		add(simulator, BorderLayout.WEST);

		analysisResultPanel = new Panel();
		analysisResultPanel.setLayout(new GridLayout(1, 1));

		analysisResultPanel.add(analysisResult = new TextArea("", 10, 10));
		analysisResult.setEditable(false);
		add(analysisResultPanel, BorderLayout.CENTER);

		stateChanges = new StateChangeStack();

		population = new SpeciesDatabase();
		birthEvents = new BirthEventStack();
		deathEvents = new DeathEventStack();
		outputPopulation();

		setSize(cx + 300, cy + 100);

		show();

		if (cnv.buffer == null) cnv.createBuffer();
	}

    public void createNewCells(int x, int y)
    {
		width = x;
		height = y;
		cnv.createBuffer();
		cell = new int[width][height];
		scan = new int[width][height];
		cellUpdates = new CellUpdateStack(width, height);
		clearCellularAutomata();
    }

    public void actionPerformed(ActionEvent e)
    {
		if (e.getSource() == startStop)
		{
			if (pause)
			{
				startStop.setLabel("Stop");
				pause = false;
				oneStep = false;
			}
			else
			{
				startStop.setLabel("Start");
				oneStep = true;
				// Consider this the same as "Step" button pressed
				// in order to execute reflectCellsToBuffer()
			}
		}
		else if (e.getSource() == step) oneStep = true;

		else if (e.getSource() == intervalField) setInterval();

		else if ((e.getSource() == reset) ||
			 (e.getSource() == spaceSizeField) ||
			 (e.getSource() == ancestorField)) resetRequest = true;
    }

    public void setInterval()
    {
		int tempInterval;

		try
		{
			tempInterval = Integer.parseInt(intervalField.getText());
		}
		catch (Exception ex)
		{
			System.out.println("Illegal interval specified.");
			intervalField.setText("10");
			tempInterval = 10;
		}
		if (tempInterval < 1)
		{
			System.out.println("Non-positive interval specified.");
			intervalField.setText("1");
			tempInterval = 1;
		}
		interval = tempInterval;
    }

    public void resetSimulation()
    {
		int spaceSize;
		String imageFile;
		File tempFile;
                boolean b = false;

		time = 0;
		timeDisplay.setText("" + time);

		setInterval();

		startStop.setLabel("Start");
		pause = true;
		oneStep = false;
		resetRequest = false;

		try
		{
			spaceSize = Integer.parseInt(spaceSizeField.getText());
		}
		catch (Exception ex)
		{
			System.out.println("Illegal space size specified.");
			spaceSizeField.setText("200");
			spaceSize = 200;
		}
		if (spaceSize < 20)
		{
			System.out.println("Too small space size specified.");
			spaceSizeField.setText("20");
			spaceSize = 20;
		}
		else if (spaceSize > 2000)
		{
			System.out.println("Too large space size specified.");
			spaceSizeField.setText("2000");
			spaceSize = 2000;
		}
		width = height = spaceSize;
		createNewCells(width, height);
		stateChanges.clear();
		population.clear();
		birthEvents.clear();
		deathEvents.clear();

		imageFile = ancestorField.getText();
                imageFile = "src/" + imageFile;
		tempFile = new File(imageFile);
		if (!tempFile.exists())
		{
			System.out.println("File \"" + imageFile + "\" does not exist.");
			ancestorField.setText("13-2.img");
			imageFile = ancestorField.getText();
		}
                
                while(b == false)
                {
                    Random rd = new Random();
                    int xxx = rd.nextInt(width);
                    int yyy = rd.nextInt(height);
                    b = putAncestor(imageFile, xxx, yyy);
                }
                
                b = false;
                
                imageFile = "src/13-2.img";
		tempFile = new File(imageFile);
		if (!tempFile.exists())
		{
			System.out.println("File \"" + imageFile + "\" does not exist.");
			ancestorField.setText("13-2.img");
			imageFile = ancestorField.getText();
		}      
                
                while(b == false)
                {
                    Random rd = new Random();
                    int xxx = rd.nextInt(width);
                    int yyy = rd.nextInt(height);
                    b = putAncestor(imageFile, xxx, yyy);
                }
                
		reflectCellsToBuffer();
		outputPopulation();                
    }

    public boolean putAncestor(String imageFile, int x, int y)
    {
		String buf;
                boolean bb = false;
		int wd, ht, len, xx, yy, i;
		char data;
		StateChange sc;

		// Measuring the size of image

		wd = ht = 0;
		try
		{
			BufferedReader br = new BufferedReader(new FileReader(imageFile));
			while ((buf = br.readLine()) != null)
			{
				ht ++;
				len = buf.length();
				if (len > wd) wd = len;
			}
			br.close();
		}
		catch (Exception ex)
		{
			System.out.println("Error: " + ex);
			System.exit(1);
		}

		// Loading image

		x = x - wd / 2;
		y = y - ht / 2;

		if ((x < 0) || (x + wd > width) || (y < 0) || (y + ht > height))
		{
			System.out.println("Image file \"" + imageFile + "\" sticks out of the space.");
//			System.exit(1);
                        return false;
		}

		xx = x;
		yy = y;

		try
		{
			BufferedReader br = new BufferedReader(new FileReader(imageFile));
			while ((buf = br.readLine()) != null)
			{
				for (i = 0, xx = x; i < buf.length(); i ++, xx ++)
				{
					data = buf.charAt(i);
					if ((data >= '1') && (data <= '9'))
						stateChanges.push(xx, yy, (int) buf.charAt(i) - (int) '0');
					else if ((data != ' ') && (data != '0'))
					{
						System.out.println("Illegal state appeared in \"" + imageFile + "\".");
						System.exit(1);
					}
				}
				yy ++;
			}
			br.close();
                        bb = true;
		}
		catch (Exception ex)
		{
			System.out.println("Error: " + ex);
			System.exit(1);
		}

		while (!stateChanges.empty())
		{
			sc = stateChanges.pop();

			x = sc.x;
			y = sc.y;
			cell[x][y] = sc.nextState;

			cellUpdates.push(x, y);
			cellUpdates.push(x > 0 ? x - 1 : width - 1, y);
			cellUpdates.push(x < width - 1 ? x + 1 : 0, y);
			cellUpdates.push(x, y > 0 ? y - 1 : height - 1);
			cellUpdates.push(x, y < height - 1 ? y + 1 : 0);
		}
                return bb;
    }

    public int getCellState(int x, int y)
    {
		if (x < 0) x = width - 1;
		if (x > width - 1) x = 0;
		if (y < 0) y = height - 1;
		if (y > height - 1) y = 0;
		return cell[x][y];
    }

    public int getScanState(int x, int y)
    {
		if (x < 0) x = width - 1;
		if (x > width - 1) x = 0;
		if (y < 0) y = height - 1;
		if (y > height - 1) y = 0;
		return scan[x][y];
    }

    public void clearCellularAutomata()
    {
		int x, y;
		for (x = 0; x < width; x ++)
		{
			for (y = 0; y < height; y ++)
			{
				cell[x][y] = QUIESCENT;
				scan[x][y] = NOT_OCCUPIED;
			}
		}
		stateChanges.clear();
		cellUpdates.clear();
    }

    public void reflectCellsToBuffer()
    {
		int x, y;
		cnv.buffer_g.setColor(cellColor[QUIESCENT]);
		cnv.buffer_g.fillRect(0, 0, width, height);
		for (x = 0; x < width; x ++)
		{
			for (y = 0; y < height; y ++)
			{
				if (scan[x][y] != NOT_OCCUPIED)
				{// indicating living loops
					cnv.buffer_g.setColor(scanColor[scan[x][y] % scanColor.length]);
					cnv.buffer_g.fillRect(x, y, 1, 1);
				}
				else if (cell[x][y] != QUIESCENT)
				{
					cnv.buffer_g.setColor(cellColor[cell[x][y]]);
					cnv.buffer_g.fillRect(x, y, 1, 1);
				}
			}
		}
		cnv.repaint();
    }

    public void outputPopulation()
    {
		SpeciesPop sppop;
		analysisResult.setText("ID: size: species: count: sequence:  \n");

		String s4 = new String();
		String s5 = new String();
		String s6 = new String();
		String s7 = new String();
                
                Integer t;
                
                int nb_sex_gene=0, pop_tot=0;
                double percent_sex=0;

		for (int i = 0; i < population.length(); i ++)
		{
			sppop = population.speciesPopOf(i);
			if (sppop.pop > 0)
			{
/********************** write simulation results in text files************************************/
//                            
//				try
//				{
//					out = new BufferedWriter(new FileWriter("Results/ID.txt",true));
//					Integer t = new Integer(i);
//					String s= t.toString();
//					out.write(s);
//					out.newLine();
//					out.close();
//				}catch (IOException e) {System.out.println("error open ID file");}
//
//
//				try
//				{
//					out = new BufferedWriter(new FileWriter("Results/size.txt",true));
//					out.write(sppop.sp.pheno.width + "x" +sppop.sp.pheno.height);
//					out.newLine();
//					out.close();
//				}catch (IOException e) {System.out.println("error open size file");}
//
//				try
//				{
//					out = new BufferedWriter(new FileWriter("Results/sequence.txt",true));
//					out.write(sppop.sp.geno.sequence);
//					out.newLine();
//					out.close();
//				}catch (IOException e) {System.out.println("error open sequence file");}
////
//				try
//				{
//					out = new BufferedWriter(new FileWriter("Results/number.txt",true));
//					Integer t2 = new Integer(sppop.pop);
                                        pop_tot = pop_tot + sppop.pop;
                                        
                                        if(sppop.sp.geno.sequence.contains("A") == true)
                                            nb_sex_gene = nb_sex_gene + sppop.pop;
//					String s2= t2.toString();
//					out.write(s2);
//					out.newLine();
//					out.close();
//				}catch (IOException e) {System.out.println("error open number file");}
//
//				try
//				{
//					out = new BufferedWriter(new FileWriter("Results/time.txt",true));
//					Long t3 = new Long(time);
//					String s3= t3.toString();
//					out.write(s3);
//					out.newLine();
//					out.close();
//				}catch (IOException e) {System.out.println("error open time file");}
//
//				try
//				{
					int length = sppop.sp.geno.sequence.length();
					int count = 0, count2 = 0, index = 0;

					while (index < length)
					{
						if (sppop.sp.geno.sequence.charAt(index) == 'G')
						{
							count = count + 1;
						}
						index = index + 1;
					}
//
//					out = new BufferedWriter(new FileWriter("Results/species.txt",true));
					t = new Integer(count);
					s4 = t.toString();
//					out.write(s4);
//					out.newLine();
//					out.close();
//				}catch (IOException e) {System.out.println("error open species file");}
//
//				try
//				{
//					out = new BufferedWriter(new FileWriter("Results/sexArm.txt",true));
//					t = new Integer(nbReproSexArm);
//					s5 = t.toString();
//					out.write(s5);
//					out.newLine();
//					out.close();
//				}catch (IOException e) {System.out.println("error open sexArm file");}
//
//				try
//				{
//					out = new BufferedWriter(new FileWriter("Results/sexLoop.txt",true));
//					t = new Integer(nbReproSexLoop);
//					s6 = t.toString();
//					out.write(s6);
//					out.newLine();
//					out.close();
//				}catch (IOException e) {System.out.println("error open sexLoop file");}
//
//				try
//				{
//					out = new BufferedWriter(new FileWriter("Results/AsexualLoop.txt",true));
//					t = new Integer(nbReproAsexualLoop);
//					s7 = t.toString();
//					out.write(s7);
//					out.newLine();
//					out.close();
//				}catch (IOException e) {System.out.println("error open AsexualLoop file");}
//                             

/**********************END  write simulation results in text files*****************************/

				analysisResult.append("" +
				i + " : " +
				sppop.sp.pheno.width + "x" +
				sppop.sp.pheno.height + " : " +
                                s4+ " : "+
                                sppop.pop + " : " +
				sppop.sp.geno.sequence + "\n");
			}
		}
                if(pop_tot > 0)
                    percent_sex = (double)nb_sex_gene/(double)pop_tot;
                else
                    percent_sex=0;
                
                percent_sex = 100*percent_sex;
                analysisResult.append("\n % loops with sex genes: " + (int)percent_sex + "\n");   
                
//                try
//                {
//                        out = new BufferedWriter(new FileWriter("Results/percent_sex_gene.txt",true));
//                        t = new Integer((int)percent_sex);
//                        s7 = t.toString();
//                        out.write(s7);
//                        out.newLine();
//                        out.close();
//                }catch (IOException e) {System.out.println("error open nb_sex_gene file:" + e);}
                
                pop_tot = 0;
                nb_sex_gene = 0;
		nbReproSexArm = 0;
		nbReproSexLoop = 0;
		nbReproAsexualLoop = 0;
    }

/*****************************NEW function to check if a loop is already having sex*******************/
    //in new F-sexyloop, condition IF done on state A (=10), not state 9
    //check if more than 1 connection is being dissolved by looking at the outside border for a a state 10
    public int checkSex(int x, int y, int dx, int dy)
    {
			int xx, yy, dxx, dyy, tempdx, NBsex;

			xx = x;
			yy = y;
			dxx = dx;
			dyy = dy;
			NBsex = 0;

			while (true)
			{
				if(getCellState(xx, yy) == 10) NBsex ++;
                                
                                //if(getCellState(xx, yy) == 9) NBsex ++;   SEXYLOOP V1
				xx += dx; if (xx < 0) xx = width - 1; if (xx > width - 1) xx = 0;
				yy += dy; if (yy < 0) yy = height - 1; if (yy > height - 1) yy = 0;
                                
                                int cc = getCellState(xx + dy, yy - dx);
				if (cc != 2  && cc != 10) break;
			}

			tempdx = dx;
			dx = dy;
			dy = -tempdx;

			while (true)
			{
				if(getCellState(xx, yy) == 10) NBsex ++;
                                
                                //if(getCellState(xx, yy) == 9) NBsex ++;   SEXYLOOP V1
				xx += dx; if (xx < 0) xx = width - 1; if (xx > width - 1) xx = 0;
				yy += dy; if (yy < 0) yy = height - 1; if (yy > height - 1) yy = 0;
                                
                                int cc = getCellState(xx + dy, yy - dx);
				if (cc != 2  && cc != 10) break;                                
			}

			tempdx = dx;
			dx = dy;
			dy = -tempdx;

			while (true)
			{
				if(getCellState(xx, yy) == 10) NBsex ++;
                                
                                //if(getCellState(xx, yy) == 9) NBsex ++;   SEXYLOOP V1
				xx += dx; if (xx < 0) xx = width - 1; if (xx > width - 1) xx = 0;
				yy += dy; if (yy < 0) yy = height - 1; if (yy > height - 1) yy = 0;
                                
                                int cc = getCellState(xx + dy, yy - dx);
				if (cc != 2  && cc != 10) break;                                                                
			}

			tempdx = dx;
			dx = dy;
			dy = -tempdx;

			while (true)
			{
				if(getCellState(xx, yy) == 10) NBsex ++;
                                //if(getCellState(xx, yy) == 9) NBsex ++;   SEXYLOOP V1
				xx += dx; if (xx < 0) xx = width - 1; if (xx > width - 1) xx = 0;
				yy += dy; if (yy < 0) yy = height - 1; if (yy > height - 1) yy = 0;
                                
                                int cc = getCellState(xx + dy, yy - dx);
				if (cc != 2  && cc != 10) break;                                                                                                
			}

		return NBsex;
	}

    public void updateCellularAutomata(int[][][][][] rule)
    {
		int x, y, c, t, r, b, l, ns, len;
		CellUpdate cu;
		StateChange sc;
		BirthEvent be;
		DeathEvent de;
		int deadID;

		while (!cellUpdates.empty())
		{
			cu = cellUpdates.pop();

			x = cu.x;
			y = cu.y;

			c = cell[x][y];
			t = getCellState(x, y - 1);
			r = getCellState(x + 1, y);
			b = getCellState(x, y + 1);
			l = getCellState(x - 1, y);
			ns = rule[c][t][r][b][l];

			if (c != ns) stateChanges.push(x, y, ns);

/*****************************MODIFIED and NEW birth events for sex detection**************************/

			if ((c == 0) && (ns == 6))
			{
				if ((t==1) && (r==2) && (b==5) && (l==2))
				{
					if (checkSex(x, y < height-1 ? y+1 : 0 , 0, 1) == 0)
					{	birthEvents.push(x, y < height-1 ? y+1 : 0 , 0, 1, false);}
				}
				else if ((t==2) && (r==5) && (b==2) && (l==1))
				{
					if (checkSex(x < width-1 ? x+1 : 0, y, 1, 0) == 0)
					{	birthEvents.push(x < width-1 ? x+1 : 0, y, 1, 0, false);}
				}
				else if ((t==5) && (r==2) && (b==1) && (l==2))
				{
					if (checkSex(x, y > 0 ? y-1 : height-1, 0, -1) == 0)
					{	birthEvents.push(x, y > 0 ? y-1 : height-1, 0, -1, false);}
				}
				else if ((t==2) && (r==1) && (b==2) && (l==5))
				{
					if (checkSex(x > 0 ? x-1 : width-1, y, -1, 0) == 0)
					{	birthEvents.push(x > 0 ? x-1 : width-1, y, -1, 0, false);}
				}
			}
/**********************************************************new SEXYLOOP***************************************************************/                        
                        //sex into loop
			if ((c == 10) && (ns == 1))
			{
				if ((t==2) && (r==1) && (b==6) && (l==1))
				{
					while (true)
					{
						x -= 1; if (x < 0) x = width - 1; if (x > width - 1) x = 0;
                                                
                                                int cc = getCellState(x, y - 1);
						if (cc != 2 && cc != 10)break;
					}
					if (checkSex(x, y, 1, 0) < 2)  //in sexyloop: <=2
					{
						deathEvents.push(x < width-1 ? x+1 : 0, y > 0 ? y-1 : height-1);
						birthEvents.push(x, y, 1, 0, true);
					}
				}
				else if ((t==1) && (r==6) && (b==1) && (l==2))
				{
					while (true)
					{
						y += 1; if (y < 0) y = height - 1; if (y > height - 1) y = 0;
                                                
                                                int cc = getCellState(x - 1, y);
						if (cc != 2 && cc != 10)break;                                                
					}
					if (checkSex(x, y , 0, -1) < 2)
					{
						deathEvents.push(x > 0 ? x-1 : width-1, y > 0 ? y-1 : height-1);
						birthEvents.push(x, y , 0, -1, true);
					}
				}
				else if ((t==6) && (r==1) && (b==2) && (l==1))
				{
					while (true)
					{
						x += 1; if (x < 0) x = width - 1; if (x > width - 1) x = 0;
                                                
                                                int cc = getCellState(x, y + 1);
						if (cc != 2 && cc != 10)break;                                                                                                
					}
					if (checkSex(x, y, -1, 0) < 2)
					{
						deathEvents.push(x > 0 ? x-1 : width-1, y < height-1 ? y+1 : 0);
						birthEvents.push(x, y, -1, 0, true);
					}
				}
				else if ((t==1) && (r==2) && (b==1) && (l==6))
				{
					while (true)
					{
						y -= 1; if (y < 0) y = height - 1; if (y > height - 1) y = 0;

                                                int cc = getCellState(x + 1, y);
						if (cc != 2 && cc != 10)break;                                                                                                                                                
					}
					if (checkSex(x, y , 0, 1) < 2)
					{
						deathEvents.push(x < width-1 ? x+1 : 0, y < height-1 ? y+1 : 0);
						birthEvents.push(x, y , 0, 1, true);
					}
				}
			}
/**********************************************************SEXYLOOP M1 and M2***************************************************************/                        
//			if ((c == 9) && (ns == 1))
//			{
//				if ((t==2) && (r==1) && (b==6) && (l==9))
//				{
//					while (true)
//					{
//						x -= 1; if (x < 0) x = width - 1; if (x > width - 1) x = 0;
//						if (getCellState(x, y - 1) != 2)break;
//					}
//					if (checkSex(x, y, 1, 0) <= 2)
//					{
//						deathEvents.push(x < width-1 ? x+1 : 0, y > 0 ? y-1 : height-1);
//						birthEvents.push(x, y, 1, 0, true);
//					}
//				}
//				else if ((t==1) && (r==6) && (b==9) && (l==2))
//				{
//					while (true)
//					{
//						y += 1; if (y < 0) y = height - 1; if (y > height - 1) y = 0;
//						if (getCellState(x - 1, y) != 2) break;
//					}
//					if (checkSex(x, y , 0, -1) <= 2)
//					{
//						deathEvents.push(x > 0 ? x-1 : width-1, y > 0 ? y-1 : height-1);
//						birthEvents.push(x, y , 0, -1, true);
//					}
//				}
//				else if ((t==6) && (r==9) && (b==2) && (l==1))
//				{
//					while (true)
//					{
//						x += 1; if (x < 0) x = width - 1; if (x > width - 1) x = 0;
//						if (getCellState(x, y + 1) != 2) break;
//					}
//					if (checkSex(x, y, -1, 0) <= 2)
//					{
//						deathEvents.push(x > 0 ? x-1 : width-1, y < height-1 ? y+1 : 0);
//						birthEvents.push(x, y, -1, 0, true);
//					}
//				}
//				else if ((t==9) && (r==2) && (b==1) && (l==6))
//				{
//					while (true)
//					{
//						y -= 1; if (y < 0) y = height - 1; if (y > height - 1) y = 0;
//						if (getCellState(x + 1, y) != 2) break;
//					}
//					if (checkSex(x, y , 0, 1) <= 2)
//					{
//						deathEvents.push(x < width-1 ? x+1 : 0, y < height-1 ? y+1 : 0);
//						birthEvents.push(x, y , 0, 1, true);
//					}
//				}
//			}
//                      //sex into arm
//			if ((c == 2) && (ns == 2) && (t==9) && (r==2) && (b==6) && (l==2))
//			{
//				y -= 1; if (y < 0) y = height - 1; if (y > height - 1) y = 0;
//				while (true)
//				{
//					x -= 1; if (x < 0) x = width - 1; if (x > width - 1) x = 0;
//					if (getCellState(x, y - 1) != 2) break;
//				}
//				if (checkSex(x, y, 1, 0) <= 2)
//				{
//					deathEvents.push(x < width-1 ? x+1 : 0, y > 0 ? y-1 : height-1);
//					birthEvents.push(x, y, 1, 0, true);
//				}
//			}
//			else if ((c == 2) && (ns == 2) && (t==2) && (r==6) && (b==2) && (l==9))
//			{
//				x -= 1; if (x < 0) x = width - 1; if (x > width - 1) x = 0;
//				while (true)
//				{
//					y += 1; if (y < 0) y = height - 1; if (y > height - 1) y = 0;
//					if (getCellState(x-1, y) != 2) break;
//				}
//				if (checkSex(x, y, 0, -1) <= 2)
//				{
//					deathEvents.push(x > 0 ? x-1 : width-1,y > 0 ? y-1 : height-1);
//					birthEvents.push(x, y, 0, -1, true);
//				}
//			}
//			else if ((c == 2) && (ns == 2) && (t==6) && (r==2) && (b==9) && (l==2))
//			{
//				y += 1; if (y < 0) y = height - 1; if (y > height - 1) y = 0;
//				while (true)
//				{
//					x += 1; if (x < 0) x = width - 1; if (x > width - 1) x = 0;
//					if (getCellState(x, y+1) != 2) break;
//				}
//				if (checkSex(x, y, -1, 0) <= 2)
//				{
//					deathEvents.push(x > 0 ? x-1 : width-1, y < height-1 ? y+1 : 0);
//					birthEvents.push(x, y, -1, 0, true);
//				}
//			}
//			else if ((c == 2) && (ns == 2) && (t==2) && (r==9) && (b==2) && (l==6))
//			{
//				x += 1; if (x < 0) x = width - 1; if (x > width - 1) x = 0;
//				while (true)
//				{
//					y -= 1; if (y < 0) y = height - 1; if (y > height - 1) y = 0;
//					if (getCellState(x + 1, y) != 2) break;
//				}
//				if (checkSex(x, y, 0, 1) <= 2)
//				{
//					deathEvents.push(x < width-1 ? x+1 : 0, y < height-1 ? y+1 : 0);
//					birthEvents.push(x, y, 0, 1, true);
//				}
//			}
/***********************END  birth events for sex detection************************************/

			if ((c == 2) && (ns == 0)&& (scan[x][y] != NOT_OCCUPIED))
			{deathEvents.push(x, y);}

		}

		while (!stateChanges.empty())
		{
			sc = stateChanges.pop();

			x = sc.x;
			y = sc.y;
			cell[x][y] = sc.nextState;

			cellUpdates.push(x, y);
			cellUpdates.push(x > 0 ? x - 1 : width - 1, y);
			cellUpdates.push(x < width - 1 ? x + 1 : 0, y);
			cellUpdates.push(x, y > 0 ? y - 1 : height - 1);
			cellUpdates.push(x, y < height - 1 ? y + 1 : 0);
		}

		while (!deathEvents.empty())
		{
			de = deathEvents.pop();
			if ((deadID = scan[de.x][de.y]) != NOT_OCCUPIED)
			{
				population.death(deadID);
				Stack st = new Stack();
				st.push(de);
				while(!st.empty())
				{
					de = (DeathEvent) st.pop();
					scan[de.x][de.y] = NOT_OCCUPIED;

					if (getScanState(de.x, de.y - 1) != NOT_OCCUPIED)
					{	st.push(new DeathEvent(de.x, de.y > 0 ? de.y-1 : height-1));}

					if (getScanState(de.x + 1, de.y) != NOT_OCCUPIED)
					{	st.push(new DeathEvent(de.x < width-1 ? de.x+1 : 0, de.y));}

					if (getScanState(de.x, de.y + 1) != NOT_OCCUPIED)
					{	st.push(new DeathEvent(de.x, de.y < height-1 ? de.y+1 : 0));}

					if (getScanState(de.x - 1, de.y) != NOT_OCCUPIED)
					{	st.push(new DeathEvent(de.x > 0 ? de.x-1 : width-1, de.y));}
				}
			}
		}

		while (!birthEvents.empty())
		{
			be = birthEvents.pop();

			int xx, yy, dx, dy, tempdx, w, h, wcheck, hcheck;
			boolean xFlag;
			String seq = "";

			xx = be.x;
			yy = be.y;
			dx = be.dx;
			dy = be.dy;
			xFlag = be.sexFlag;

			w = 0;
			while (true)
			{
				seq += ("" + cell[xx][yy]);
				w ++;
				xx += dx; if (xx < 0) xx = width - 1; if (xx > width - 1) xx = 0;
				yy += dy; if (yy < 0) yy = height - 1; if (yy > height - 1) yy = 0;
                                int cc = getCellState(xx + dy, yy - dx);
				if (cc != 2 && cc != 10) break;
			}
			w --;

			tempdx = dx;
			dx = dy;
			dy = -tempdx;

			h = 0;
			while (true)
			{
				seq += ("" + cell[xx][yy]);
				h ++;
				xx += dx; if (xx < 0) xx = width - 1; if (xx > width - 1) xx = 0;
				yy += dy; if (yy < 0) yy = height - 1; if (yy > height - 1) yy = 0;
                                int cc = getCellState(xx + dy, yy - dx);
				if (cc != 2) break;                                
			}
			h --;

			tempdx = dx;
			dx = dy;
			dy = -tempdx;

			wcheck = 0;
			while (true)
			{
				seq += ("" + cell[xx][yy]);
				wcheck ++;
				xx += dx; if (xx < 0) xx = width - 1; if (xx > width - 1) xx = 0;
				yy += dy; if (yy < 0) yy = height - 1; if (yy > height - 1) yy = 0;
                                int cc = getCellState(xx + dy, yy - dx);
				if (cc != 2) break;                                                                
			}
			wcheck --;

			tempdx = dx;
			dx = dy;
			dy = -tempdx;

			hcheck = 0;
			while (true)
			{
				seq += ("" + cell[xx][yy]);
				hcheck ++;
				xx += dx; if (xx < 0) xx = width - 1; if (xx > width - 1) xx = 0;
				yy += dy; if (yy < 0) yy = height - 1; if (yy > height - 1) yy = 0;
                                int cc = getCellState(xx + dy, yy - dx);
				if (cc != 2 && cc != 10) break;                                                                                                
			}
			hcheck --;

			Species sp = new Species(w, h, seq);

			if ((w == wcheck) && (h == hcheck) && (sp.geno.sequence != null) && (w >= 2) && (h >= 2))
			{
				if(xFlag == true)
				{	nbReproSexLoop++;}
				else
				{	nbReproAsexualLoop++;}

//				System.out.println("New specie");
//				System.out.println(w);
//				System.out.println(wcheck);
//				System.out.println(h);
//				System.out.println(hcheck);
//				System.out.println(seq);
//				System.out.println(sp.geno.sequence);
//				System.out.println("\n");

				population.birth(sp);
				int bornID = population.indexOf(sp);

				xx -= dx; if (xx < 0) xx = width - 1; if (xx > width - 1) xx = 0;
				yy -= dy; if (yy < 0) yy = height - 1; if (yy > height - 1) yy = 0;

				tempdx = dx;
				dx = dy;
				dy = -tempdx;

				xx += dx; if (xx < 0) xx = width - 1; if (xx > width - 1) xx = 0;
				yy += dy; if (yy < 0) yy = height - 1; if (yy > height - 1) yy = 0;

				int corner = 0;
				while(corner < 4)
				{
					scan[xx][yy] = bornID;
					if (getCellState(xx + dx, yy + dy) != 2)
					{
						tempdx = dx;
						dx = dy;
						dy = -tempdx;
						corner ++;
					}
					else
					{
						xx += dx; if (xx < 0) xx = width - 1; if (xx > width - 1) xx = 0;
						yy += dy; if (yy < 0) yy = height - 1; if (yy > height - 1) yy = 0;
					}
				}
			}
			else
			{
				if(xFlag == true)
				{	nbReproSexArm++;}
			}
		}
    }

    public void simulation(int[][][][][] rule)
    {
//        String myNewDir = "Results";
//        new File(myNewDir).mkdirs();
        
		while (time<1000000)
		{
			if (resetRequest) resetSimulation();
			if ((!pause) || oneStep)
			{
				updateCellularAutomata(rule);
				time ++;
				timeDisplay.setText("" + time);
				if (oneStep || (time % interval == 0))
				{
					reflectCellsToBuffer();
					outputPopulation();
				}
				if (oneStep)
				{
					startStop.setLabel("Start");
					pause = true;
					oneStep = false;
				}
			}
		}
    }
}


// For CA updating

class StateChange
{
    public int x, y;
    public int nextState;

    StateChange(int xx, int yy, int nn)
    {
		x = xx;
		y = yy;
		nextState = nn;
    }
}


class StateChangeStack
{
    private Stack changeStack;

    StateChangeStack()
    {
		changeStack = new Stack();
		clear();
    }

    public void clear()
    {
		changeStack.clear();
    }

    public boolean empty()
    {
		return changeStack.empty();
    }

    public void push(int x, int y, int nextState)
    {
		changeStack.push(new StateChange(x, y, nextState));
    }

    public StateChange pop()
    {
		if (changeStack.empty()) return null;
		else return (StateChange) changeStack.pop();
    }
}


class CellUpdate
{
    public int x, y;

    CellUpdate(int xx, int yy)
    {
		x = xx;
		y = yy;
    }
}

class CellUpdateStack
{
    private Stack updateStack;
    private boolean[][] inStack;
    private int spacex, spacey;

    CellUpdateStack(int x, int y)
    {
		updateStack = new Stack();
		inStack = new boolean[x][y];
		spacex = x;
		spacey = y;
		clear();
    }

    public void clear()
    {
		updateStack.clear();
		for (int i = 0; i < spacex; i ++)
			for (int j = 0; j < spacey; j ++)
			inStack[i][j] = false;
    }

    public boolean empty()
    {
		return updateStack.empty();
    }

    public void push(int x, int y)
    {
		if (!inStack[x][y])
		{
			inStack[x][y] = true;
			updateStack.push(new CellUpdate(x, y));
		}
    }

    public CellUpdate pop()
    {
		CellUpdate cu;

		if (updateStack.empty()) return null;
		else
		{
			cu = (CellUpdate) updateStack.pop();
			inStack[cu.x][cu.y] = false;
			return cu;
		}
    }
}


// For maintaining species database and birth/death detection

class Phenotype
{
    public int width, height;

    Phenotype(int w, int h)
    {
		width = w;
		height = h;
    }

    public boolean equals(Phenotype another)
    {
		if ((width == another.width) && (height == another.height)) return true;
		else return false;
    }
}

class Genotype
{
    public String rawData;
    public String sequence;

    Genotype(String raw)
    {
		rawData = raw;
		compressRawData();
    }

    public boolean equals(Genotype another)
    {
		if ((sequence == null) || (another.sequence == null)) return false;
		if (sequence.equals(another.sequence)) return true;
		else return false;
	}

/**********************MODIFIED function to compress as well the data when sex append******************/
    private void compressRawData()
    {
		int i;

		boolean b = false, bT = false, bG = false, bA = false;

		sequence = "";
		for (i = 0; i < rawData.length(); i ++)
		{
			if (rawData.regionMatches(i, "1", 0, 1) && i == 0)
			{ b = true;}
			else if (rawData.regionMatches(i, "41", 0, 2) && i == 0)
			{
				bT = true;
				i += 1;
			}
			else if (rawData.regionMatches(i, "71", 0, 2) && i == 0)
			{
				bG = true;
				i += 1;
			}
			else if (rawData.regionMatches(i, "91", 0, 2) && i == 0)
			{
				bA = true;
				i += 1;
			}                        
			else if (rawData.regionMatches(i, "1", 0, 1))
			{
				sequence += "C";
				if (i == (rawData.length()-1) && b == true)
				{	sequence += "C";}
			}
			else if (rawData.regionMatches(i, "041", 0, 3))
			{
				sequence += "T";
				i += 2;
			}
			else if (rawData.regionMatches(i, "071", 0, 3))
			{
				sequence += "G";
				i += 2;
			}
			else if (rawData.regionMatches(i, "091", 0, 3))
			{
				sequence += "A";
				i += 2;
			}                        
			else if(rawData.regionMatches(i, "04", 0, 2) && b == true)
			{
				sequence += "T";
				i += 1;
			}
			else if(rawData.regionMatches(i, "07", 0, 2) && b == true)
			{
				sequence += "G";
				i += 1;
			}
			else if(rawData.regionMatches(i, "09", 0, 2) && b == true)
			{
				sequence += "A";
				i += 1;
			}                        
			else if (rawData.regionMatches(i, "0", 0, 1))
			{
				if(bT == true)
				{	sequence += "T";}
				if(bG == true)
				{	sequence += "G";}
				if(bA == true)
				{	sequence += "A";}                                

				i += 1;
			}
			else
			{
				sequence = null;
				break;
			}
		}
    }
}


class Species
{
    public Phenotype pheno;
    public Genotype geno;

    Species (Phenotype p, Genotype g)
    {
		pheno = p;
		geno = g;
    }

    Species (int w, int h, String raw)
    {
		pheno = new Phenotype(w, h);
		geno = new Genotype(raw);
    }

    public boolean equals(Species another)
    {
		if ((pheno.equals(another.pheno)) && (geno.equals(another.geno))) return true;
		else return false;
    }
}


class SpeciesPop
{
    public Species sp;
    public int pop;

    SpeciesPop(Species s, int p)
    {
		sp = s;
		pop = p;
    }
}


class SpeciesDatabase
{
    private ArrayList database;

    SpeciesDatabase()
    {
		database = new ArrayList();
		clear();
    }

    public void clear()
    {
		database.clear();
    }

    public int length()
    {
		return database.size();
    }

    public SpeciesPop speciesPopOf(int i)
    {
		if ((i < 0) || (i >= database.size())) return null;
		else return (SpeciesPop) database.get(i);
    }

    public int indexOf(Species s)
    {
		int i;

		for (i = 0; i < database.size(); i ++)
			if (s.equals(((SpeciesPop) database.get(i)).sp)) return i;

		database.add(new SpeciesPop(s, 0));
		return i;
    }

    public void birth(Species s)
    {
		SpeciesPop sppop = null;

		for (int i = 0; i < database.size(); i ++)
		{
			sppop = (SpeciesPop) database.get(i);
			if (s.equals(sppop.sp)) break;
			sppop = null;
		}

		if (sppop == null) database.add(sppop = new SpeciesPop(s, 0));

		sppop.pop ++;
    }

    public void death(int i)
    {
		SpeciesPop sppop = (SpeciesPop) database.get(i);
		sppop.pop --;
		if (sppop.pop < 0)
		{
			System.out.println("Non-existent species died!?");
			System.exit(1);
		}
    }
}


class BirthEvent
{
    public int x, y, dx, dy;
    boolean sexFlag;

    BirthEvent(int xx, int yy, int ddx, int ddy, boolean xFlag)
    {
		x = xx;
		y = yy;
		dx = ddx;
		dy = ddy;
		sexFlag = xFlag;
    }
}


class BirthEventStack
{
    private Stack eventStack;

    BirthEventStack()
    {
		eventStack = new Stack();
		clear();
    }

    public void clear()
    {
		eventStack.clear();
    }

    public boolean empty()
    {
		return eventStack.empty();
    }

    public void push(int x, int y, int dx, int dy, boolean sexFlag)
    {
		eventStack.push(new BirthEvent(x, y, dx, dy, sexFlag));
    }

    public BirthEvent pop()
    {
		if (eventStack.empty()) return null;
		else return (BirthEvent) eventStack.pop();
    }
}


class DeathEvent
{
    public int x, y;

    DeathEvent(int xx, int yy)
    {
		x = xx;
		y = yy;
    }
}


class DeathEventStack
{
    private Stack eventStack;

    DeathEventStack()
    {
		eventStack = new Stack();
		clear();
    }

    public void clear()
    {
		eventStack.clear();
    }

    public boolean empty()
    {
		return eventStack.empty();
    }

    public void push(int x, int y)
    {
		eventStack.push(new DeathEvent(x, y));
    }

    public DeathEvent pop()
    {
		if (eventStack.empty()) return null;
		else return (DeathEvent) eventStack.pop();
    }
}


/*******************MODIFIED Main application using a 10 states' rule(0 to 9)*********************************/

class F_Sexyloop
{
    static int canvasWidth = 800;
    static int canvasHeight = 800;
    static CellularAutomata ca;
    static int[][][][][] rule = new int[12][12][12][12][12];

    public static void defineRules(String ruleFile, boolean introduceDissolver)
    {
		String buf;
		int a, b, c, d, e, f;

		for (a = 0; a <= 11; a ++)
			for (b = 0; b <= 11; b ++)
			for (c = 0; c <= 11; c ++)
				for (d = 0; d <= 11; d ++)
				for (e = 0; e <= 11; e ++)
					rule[a][b][c][d][e] = -1;

		try
		{
			BufferedReader br = new BufferedReader(new FileReader(ruleFile));
			while ((buf = br.readLine()) != null) 
                        {
                            char id = buf.charAt(0);
                            if(id == 'A')
                            {
                                a = 10;
                            }
                            else
                            {
                                if(id == 'B')
                                {
                                    a = 11;
                                }
                                else
                                    a = (int) id - (int) '0';
                            }
                            
                            id = buf.charAt(1);
                            if(id == 'A')
                            {
                                b = 10;
                            }
                            else
                            {
                                if(id == 'B')
                                {
                                    b = 11;
                                }
                                else
                                    b = (int) id - (int) '0';
                            }
                            
                            id = buf.charAt(2);
                            if(id == 'A')
                            {
                                c = 10;
                            }
                            else
                            {
                                if(id == 'B')
                                {
                                    c = 11;
                                }
                                else
                                    c = (int) id - (int) '0';
                            }
                            
                            id = buf.charAt(3);
                            if(id == 'A')
                            {
                                d = 10;
                            }
                            else
                            {
                                if(id == 'B')
                                {
                                    d = 11;
                                }
                                else
                                    d = (int) id - (int) '0';
                            }
                            
                            id = buf.charAt(4);
                            if(id == 'A')
                            {
                                e = 10;
                            }
                            else
                            {
                                if(id == 'B')
                                {
                                    e = 11;
                                }
                                else
                                    e = (int) id - (int) '0';
                            }
                            
                            id = buf.charAt(5);
                            if(id == 'A')
                            {
                                f = 10;
                            }
                            else
                            {
                                if(id == 'B')
                                {
                                    f = 11;
                                }
                                else
                                    f = (int) id - (int) '0';
                            }  

                            //create rotationaly symetric rules
                            rule[a][b][c][d][e] = f;
                            rule[a][c][d][e][b] = f;
                            rule[a][d][e][b][c] = f;
                            rule[a][e][b][c][d] = f;
			}
			br.close();
		}
		catch (Exception ex)
		{
			System.out.println("Error: " + ex);
			System.exit(1);
		}

		if (!introduceDissolver) return;

		// Defining rules related to state '8'

		for (a = 0; a <= 11; a ++)
			for (b = 0; b <= 11; b ++)
			for (c = 0; c <= 11; c ++)
				for (d = 0; d <= 11; d ++)
				for (e = 0; e <= 11; e ++)
					if (rule[a][b][c][d][e] == -1) {

					if (a == 8) rule[a][b][c][d][e] = 0;

					else if ((b == 8) || (c == 8) || (d == 8) || (e == 8))
						switch(a) {

						case 0:
						case 1:
						if (((b >= 2) && (b <= 7)) || ((c >= 2) && (c <= 7)) ||
							((d >= 2) && (d <= 7)) || ((e >= 2) && (e <= 7)))
							rule[a][b][c][d][e] = 8;
						else rule[a][b][c][d][e] = a;
						break;

						case 2:
						case 3:
						case 5:
						rule[a][b][c][d][e] = 0;
						break;

						case 4:
						case 6:
						case 7:
						rule[a][b][c][d][e] = 1;
						break;
						}
					}

		// Defining all undefined rules

		for (a = 0; a <= 11; a ++)
			for (b = 0; b <= 11; b ++)
			for (c = 0; c <= 11; c ++)
				for (d = 0; d <= 11; d ++)
				for (e = 0; e <= 11; e ++)
					if (rule[a][b][c][d][e] == -1)
					{
						if (a == 0) rule[a][b][c][d][e] = 0;
						else rule[a][b][c][d][e] = 8;
					}
    }

    public static void main (String args[])
    {
        defineRules("src/rule_sex", true);

        ca = new CellularAutomata(canvasWidth, canvasHeight);
        ca.resetSimulation();
        ca.simulation(rule);
    }
}
